/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.anfrage;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;

import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.puk.param.hierarchie.HierarchieObjekt;
import de.bsvrz.puk.param.lib.ParameterInfo;
import de.bsvrz.puk.param.lib.ParameterInfo.SerializableParameterInfo;
import de.bsvrz.puk.param.lib.ParametrierungsAnfrageTyp;
import de.bsvrz.puk.param.param.Parametrierung;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Klasse zur Ermittlung einer Antwort auf die Anfrage nach konfigurierbaren
 * Objekten.<br>
 * Parameter für eine Datenspezifikation können an verschiedenen Objekten
 * gespeichert werden, als lokale Parameter direkt an einem Objekt, an den dem
 * Objekt übergeordnetem Objekt innerhalb einer konfigurierten Objekthierarchie
 * oder am Typobjekt eines konkreten Objekts.<br>
 * Die Funktion ermittelt die Objekte, deren Parameter potentiell mit dem Setzen
 * eines Parameters an einem übergebenen Objekt beeinflusst werden können. Es
 * wird nicht überprüft, ob die untergeordneten Objekte durch lokale Daten oder
 * Daten aus einer untergeordneten Ebene des Vorgangs zur Parameterermittlung
 * bereits mit Daten versorgt sind.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
class KonfigurierbareObjekteErmittler extends ParameterAnfrageBeantworter {

	/** Logger für Debugausgaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * die Parameterspezifikation, für die die Anfrage ausgeführt werden soll.
	 */
	private final ParameterInfo info;

	/**
	 * Konstruktor, erzeugt eine Instanz des Ermittlers.<br>
	 * Das Objekt erhält das Parametrierungsmodul, aus dem die Informationen
	 * ermittelt werden sollen übergeben. Die ermittelten Ergebnisse werden an
	 * das Anfrager-Objekt zurückgesendet, welches ein Systemobjekt vom Typ
	 * "Applikation" sein sollte. Der Index wird zur Identifizierung der Anfrage
	 * auf Client-Seite wieder in der Antwort zurückgeliefert. Die Daten
	 * enthalten den Inhalt der eigentlichen Anfrage.
	 *
	 * @param parametrierung
	 *            das verwendete Parametrierungsmodul
	 * @param anfrager
	 *            das Systemobjekt, das die anfragende Applikation repräsentiert
	 * @param index
	 *            der Index der Anfrage
	 * @param daten
	 *            die Daten der Anfrage
	 */
	KonfigurierbareObjekteErmittler(final Parametrierung parametrierung, final SystemObject anfrager, final int index,
			final Object daten) {
		super(parametrierung, anfrager, index);
		if (!(daten instanceof SerializableParameterInfo)) {
			throw new IllegalArgumentException("Die Daten der Anfrage konnten nicht interpretiert werden: " + daten);
		}
		info = ((SerializableParameterInfo) daten).erzeugeParameterInfo(getParametrierung().getVerbindung());
	}

	/**
	 * Die Funktion ermittelt abhängig vom übergebenen Typ des Objekts für das
	 * die Anfrage ermittelt wurde, eine Liste der Objekte des betreffenden Typs
	 * oder der unterhalb eines Hierarchieobjektes definierten Objekte. Der
	 * Inhalt der Liste ist unabhängig vom Inhalt des konkreten
	 * Parametrierungsdatensatzes der Parametrierung, da er die potentiell
	 * parametrierenden Objekte liefern soll.<br>
	 * Die ermittelten Systemobjekte werden als Liste von Systemobjekt-ID
	 * geliefert.
	 */
	@Override
	protected Serializable ermittleAntwort() {

		KonfigurierbareObjekteErmittler.LOGGER.fine("Konfigurierbare Objekte ermittelten für", getAnfrager());

		final Collection<Long> ergebnis = new ArrayList<>();

		final SystemObject requestObject = info.getObjekt();
		if (requestObject != null) {
			if (requestObject instanceof SystemObjectType) {
				for (final SystemObject objekt : ((SystemObjectType) requestObject).getElements()) {
					ergebnis.add(objekt.getId());
				}
			} else {
				ergebnis.add(requestObject.getId());
				final HierarchieObjekt hierarchieObjekt = getParametrierung().getHierarchieManager()
						.getHierarchieObjekt(requestObject.getId());
				if (hierarchieObjekt != null) {
					if (hierarchieObjekt.getObjekt().isOfType(info.getTyp())) {
						ergebnis.add(hierarchieObjekt.getObjekt().getId());
					} else {
						for (final SystemObject objekt : hierarchieObjekt.getNachFolgerTypObjekte(info.getTyp())) {
							ergebnis.add(objekt.getId());
						}
					}
				}
			}
		}

		KonfigurierbareObjekteErmittler.LOGGER
				.fine("Es wurden " + ergebnis.size() + " konfigurierbare Objekte ermittelt");
		KonfigurierbareObjekteErmittler.LOGGER.finer("Ermittelte Objekte:", ergebnis);
		return ergebnis.toArray(new Long[ergebnis.size()]);
	}

	@Override
	protected ParametrierungsAnfrageTyp getAntwortTyp() {
		return ParametrierungsAnfrageTyp.AENDERBARE_OBJEKTE;
	}
}
