/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.anfrage;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;

import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.puk.param.lib.Parameter;
import de.bsvrz.puk.param.lib.ParameterInfo;
import de.bsvrz.puk.param.lib.ParameterInfo.SerializableParameterInfo;
import de.bsvrz.puk.param.lib.ParametrierungsAnfrageTyp;
import de.bsvrz.puk.param.param.Parametrierung;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Realisierung einer Klasse zur Ermittlung einer Antwort auf die Anfrage nach
 * einem Wurzelobjekt.<br>
 * Als Wurzelobjekt wird das Objekt betrachtet, das die Daten für eine
 * übergebene Parameterspezifikation bereitstellt.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
class SucheWurzelErmittler extends ParameterAnfrageBeantworter {

	/** Logger für Debugausgaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/** die Parameterdaten der Anfrage. */
	private final ParameterInfo[] anfrageInfos;

	/**
	 * Konstruktor, erzeugt eine Instanz des Ermittlers.<br>
	 * Die Antwort der Anfrage wird mit Hilfe der Daten des übergebenen
	 * Parametrierungsmoduls ermittelt, Die ermittelte Antwort wird an das
	 * übergebene Anfrageobjekt, welches ein Systemobjekt vom Typ "Applikation
	 * sein sollte" gesendet. Der übergebene Index wird in die Antwort
	 * eingebettet und dient auf der Seite des Clients zur Zuordnung von
	 * Anfragen zu Antworten. Als Daten für die Anfrage wird ein Feld von
	 * serialisierten Paremeterspezifikationen erwartet.
	 *
	 * @param parametrierung
	 *            das verwendete Parametrierungsmodul
	 * @param anfrager
	 *            das Systemobjekt, das die anfragende Applikation repräsentiert
	 * @param index
	 *            die Index der Anfrage
	 * @param daten
	 *            die Parameterdaten der Anfrage
	 */
	SucheWurzelErmittler(final Parametrierung parametrierung, final SystemObject anfrager, final int index,
			final Object daten) {
		super(parametrierung, anfrager, index);

		if (!(daten instanceof SerializableParameterInfo[])) {
			throw new IllegalArgumentException("Die Daten der Anfrage konnten nicht interpretiert werden: " + daten);
		}

		final SerializableParameterInfo[] anfrageFeld = (SerializableParameterInfo[]) daten;
		anfrageInfos = new ParameterInfo[((SerializableParameterInfo[]) daten).length];

		for (int idx = 0; idx < anfrageFeld.length; idx++) {
			anfrageInfos[idx] = anfrageFeld[idx].erzeugeParameterInfo(parametrierung.getVerbindung());
		}
	}

	/**
	 * für jede übergebene Parameterspezifikation wird das Objekt bestimt, in
	 * dem gültige Parameter innerhalb des Parametrierungsmoduls zur Verfügung
	 * stehen. Das Ergebnis ist ein Feld mit SystemObjekt-ID, die die
	 * ermittelten Wurzelobjekte repräsentieren. Wenn keine Parameterdaten
	 * ermittelt werden konnten, wird als ID der Wert 0 eingetragen, d.h. das
	 * Antwortfeld entspricht immer der Größe der Anfrage.
	 */
	@Override
	protected Serializable ermittleAntwort() {

		SucheWurzelErmittler.LOGGER.fine("Parameter lesen für " + getAnfrager());

		final Collection<Long> ergebnis = new ArrayList<>();
		for (final ParameterInfo info : anfrageInfos) {
			final Parameter parameter = getParametrierung().ermittleParameter(info, true);
			SystemObject wurzel = null;
			if (parameter != null) {
				wurzel = parameter.getQuelle().getObjekt();
			}
			if (wurzel == null) {
				ergebnis.add(0L);
			} else {
				ergebnis.add(wurzel.getId());
			}
		}

		SucheWurzelErmittler.LOGGER.fine("Es wurden " + ergebnis.size() + " Parameter-Wurzelobjekte ermittelt");
		SucheWurzelErmittler.LOGGER.finer("Ermittelte Objekt-ID: " + ergebnis);

		return ergebnis.toArray(new Long[ergebnis.size()]);
	}

	@Override
	protected ParametrierungsAnfrageTyp getAntwortTyp() {
		return ParametrierungsAnfrageTyp.PARAMETER_WURZEL;
	}
}
