/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.hierarchie;

import java.util.HashMap;
import java.util.Map;

import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;

/**
 * Repräsentation der konkreten Hierarchieobjekte entsprechend der
 * Hierarchiedefinition in der Konfiguration des Datenverteilers.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
class ObjektHierarchie {

	/**
	 * Klasse, die einen Thread zum Neuerstellen der Hierarchie repräsentiert.
	 */
	private static final class HierarchieRebuilder extends Thread {

		/** die Typhierarchie, für die die Objekte berechnet werden. */
		private final TypHierarchie typHierarchie;

		/** die Objekthierarchie, die aktualisiert werden soll. */
		private final ObjektHierarchie objektHierarchie;

		/**
		 * Konstruktor, erstellt einen Thread für die Aktualisierung der
		 * übergebenen Objekthierarchie auf Basis der übergebenen Typhierarchie.
		 *
		 * @param objektHierarchie
		 *            die zu aktualisierende Objekthierarchie
		 * @param typHierarchie
		 *            die verwendete Hierarchiedefinition
		 */
		HierarchieRebuilder(final ObjektHierarchie objektHierarchie, final TypHierarchie typHierarchie) {
			this.objektHierarchie = objektHierarchie;
			this.typHierarchie = typHierarchie;
		}

		@Override
		public void run() {
			synchronized (objektHierarchie.hierarchieObjekte) {
				objektHierarchie.hierarchieObjekte.clear();
				for (final TypHierarchieObjekt typObjekt : typHierarchie.getEintraege()) {
					objektHierarchie.erzeugeObjekte(typObjekt);
				}
			}
		}
	}

	/** die Menge aller Objekte der Hierarchie. */
	private final Map<Long, HierarchieObjekt> hierarchieObjekte = new HashMap<>();

	/**
	 * erzeugt ein neues Hierarchieobjekt für das übergebene
	 * Datenverteiler-Systemobjekt, falls für das Objekt noch kein solches
	 * existiert.
	 *
	 * @param objekt
	 *            das Systemobjekt
	 * @return das Hierarchieobjekt
	 */
	private HierarchieObjekt erzeugeHierarchieObject(final SystemObject objekt) {
		synchronized (hierarchieObjekte) {
			HierarchieObjekt result = hierarchieObjekte.get(objekt.getId());
			if (result == null) {
				result = new HierarchieObjekt(objekt);
				hierarchieObjekte.put(objekt.getId(), result);
			}
			return result;
		}
	}

	/**
	 * erzeugt die Nachfolgerobjekte eines Hierarchieobjekts.
	 *
	 * @param objekt
	 *            das Hierarchieobjekt, dessen Nachfolger erzeugt werden sollen
	 * @param menge
	 *            der Name der Menge, die die Nachfolgerelemente enthält
	 * @param nachfolgerTyp
	 *            der Typ der Nachfolgerobjekte
	 */
	private void erzeugeHierarchieObjektNachfolger(final HierarchieObjekt objekt, final String menge,
			final SystemObjectType nachfolgerTyp) {

		if (objekt.getObjekt() instanceof ConfigurationObject) {
			final ConfigurationObject cfgObj = (ConfigurationObject) objekt.getObjekt();
			for (final SystemObject mengenObjekt : cfgObj.getObjectSet(menge).getElements()) {
				if (mengenObjekt.isOfType(nachfolgerTyp)) {
					final HierarchieObjekt nachFolger = erzeugeHierarchieObject(mengenObjekt);
					nachFolger.setVorgaenger(objekt);
					objekt.setNachfolger(nachFolger);
				}
			}
		}
	}

	/**
	 * erzeugt alle Hierachieobjekte für das übergebene Element aus der
	 * Hierarchiedefinition.
	 *
	 * @param typObjekt
	 *            das Element der Hierarchiedefinition
	 */
	void erzeugeObjekte(final TypHierarchieObjekt typObjekt) {
		synchronized (hierarchieObjekte) {
			final SystemObjectType typ = typObjekt.getTyp();
			if (typObjekt.getNachfolgerTyp() != null) {
				for (final SystemObject object : typ.getElements()) {
					if (object instanceof ConfigurationObject) {
						erzeugeHierarchieObjektNachfolger(erzeugeHierarchieObject(object), typObjekt.getMenge(),
								typObjekt.getNachfolgerTyp());
					}
				}
			}
		}
	}

	/**
	 * liefert das Hierarchieobjekt, welches das Systemobjekt mit der
	 * übergebenen ID repräsentiert.
	 *
	 * @param id
	 *            die ID
	 * @return das Hierarchieobjekt oder <i>null</i>, wenn kein solches
	 *         existiert
	 */
	HierarchieObjekt getHierarchieObjekt(final long id) {
		synchronized (hierarchieObjekte) {
			return hierarchieObjekte.get(id);
		}
	}

	/**
	 * rekonfiguriert die Objekthierarchie auf Grund der übergebenen
	 * Typhierarchie neu.<br>
	 * Die Neuberechnung der Hierarchie erfolgt in einem gesonderten Thread, da
	 * das die Funktion als Reaktion auf das Anlegen eines neuen dynamischen
	 * Objekts erfolgt und die Kommunikation mit dem Datenverteiler nicht
	 * blockiert werden soll.
	 *
	 * @param typHierarchie
	 *            die Typhierarchie
	 */
	void rebuild(final TypHierarchie typHierarchie) {
		new HierarchieRebuilder(this, typHierarchie).start();
	}
}
