/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.param;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;

import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.puk.param.hierarchie.HierarchieObjekt;
import de.bsvrz.puk.param.lib.DatenSchreibenException;
import de.bsvrz.puk.param.lib.Parameter;
import de.bsvrz.puk.param.lib.ParameterInfo;
import de.bsvrz.puk.param.lib.QuellTyp;
import de.bsvrz.puk.param.lib.daten.DataWithTime;
import de.bsvrz.sys.funclib.bitctrl.daf.DaVKonstanten;
import de.bsvrz.sys.funclib.bitctrl.daf.DavTools;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Klasse zur Bereitstellung der Parameter für die Objekte, die durch den
 * Datensatz "atg.parametrierung" der Parametrierungsapplikation definiert sind.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
class ParameterProvider extends Thread implements ClientReceiverInterface, ClientSenderInterface, Provider {

	/** der Logger für die Ausgabe von DEBUG-Meldungen. */
	private static final Debug LOGGER = Debug.getLogger();

	/** der Aspekt für die Veröffentlichung des Sollzustandes. */
	private final Aspect sollAsp;

	/** der Aspekt zum Empfang von Vorgaben von Sollparametern. */
	private final Aspect vorgabeAsp;

	/**
	 * die Menge der versorgten Parameterspezifikationen und ihre Standardwerte.
	 */
	private final Map<ParameterInfo, Data> provided = new HashMap<>();

	/** die Parametrierung, für die die Parameter bereitgestellt werden. */
	private final Parametrierung parametrierung;

	/** Queue zur Verwaltung der Vorgaben an die Parametrierung. */
	private final Queue<ResultData> vorgaben = new LinkedList<>();

	/**
	 * Reader zum Auslesen der Standardparameter, die mit dem Aspekt
	 * "asp.parameterDefault" abgelegt sind.
	 */
	private DefaultParameterReader defaultParameterReader;

	/** Mutex zur Synchronisation der Sendeschlange. */
	private final Object senderMutex = new Object();

	/**
	 * erzeugt einen {@link ParameterProvider} für die übergebene
	 * {@link Parametrierung}.
	 *
	 * @param parametrierung
	 *            die Parametrierung
	 */
	ParameterProvider(final Parametrierung parametrierung) {
		this.parametrierung = parametrierung;

		vorgabeAsp = parametrierung.getVerbindung().getDataModel().getAspect(DaVKonstanten.ASP_PARAMETER_VORGABE);
		sollAsp = parametrierung.getVerbindung().getDataModel().getAspect(DaVKonstanten.ASP_PARAMETER_SOLL);

		if (parametrierung.isOldDefault()) {
			defaultParameterReader = new DefaultParameterReader(parametrierung);
		}
	}

	@Override
	public void connect(final ParameterInfo info) {

		if (isProvidedIdentifikation(info)) {
			synchronized (provided) {

				if (!provided.containsKey(info)) {

					DavTools.validiereSimulationsVariante(info.getSim());
					Data daten = null;

					if (defaultParameterReader != null) {
						daten = defaultParameterReader.getDaten(info);
					}
					provided.put(info, daten);

					final Parameter parameter = parametrierung.ermittleParameter(info, true);
					daten = null;
					long zeitPunkt = 0;
					if (parameter != null) {
						daten = parameter.getData();
						zeitPunkt = parameter.getZeitpunkt();
					}
					if (zeitPunkt <= 0) {
						zeitPunkt = parametrierung.getVerbindung().getTime();
					}

					try {
						parametrierung.getVerbindung().subscribeSource(this, new ResultData(info.getObjekt(),
								new DataDescription(info.getAtg(), sollAsp, info.getSim()), zeitPunkt, daten));
					} catch (final OneSubscriptionPerSendData e) {
						// sollte nicht auftreten, da die Anmeldung bereits
						// vorher geprüft wird
						throw new RuntimeException("Doppelte Anmeldung für :" + info, e);
					}

					parametrierung.getVerbindung().subscribeReceiver(this, info.getObjekt(),
							new DataDescription(info.getAtg(), vorgabeAsp, info.getSim()), ReceiveOptions.normal(),
							ReceiverRole.drain());
				}
			}
		}
	}

	@Override
	public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
		// alle Daten werden sofort veröffentlicht.
		// Ein Versand erst beim Empfang einer Sendesteuerung ist nicht möglich,
		// da ansonsten vom Empfänger zuerst ein "Keine Daten" - Datensatz
		// empfangen wird.
	}

	@Override
	public void disconnect(final ParameterInfo info, final boolean entferneDaten) {
		if (isProvidedIdentifikation(info)) {
			if (provided.containsKey(info)) {
				provided.remove(info);
				parametrierung.getVerbindung().unsubscribeReceiver(this, info.getObjekt(),
						new DataDescription(info.getAtg(), vorgabeAsp, info.getSim()));
				parametrierung.getVerbindung().unsubscribeSender(this, info.getObjekt(),
						new DataDescription(info.getAtg(), sollAsp, info.getSim()));
				if (entferneDaten) {
					try {
						parametrierung.getPersistenz().entferne(info);
					} catch (final DatenSchreibenException e) {
						LogTools.log(ParameterProvider.LOGGER, ParamMessage.ErrorSchreibenDatensatz, e.getMessage());
					}
				}
			}
		}
	}

	/**
	 * liefert die Standarddaten eines Objekts, die in der Konfiguration mit dem
	 * Aspekt "asp.parameterDefault" abgelegt sind.<br>
	 * Die Funktionalität entfällt in der endgültigen Version.
	 *
	 * @param info
	 *            die Parameterspezifikation.
	 * @return den Parameterdatensatz oder <i>null</i>, wenn keiner ermittelt
	 *         wurde
	 */
	Parameter getOldStyleDefaultDaten(final ParameterInfo info) {
		Parameter result = null;
		if (parametrierung.isOldDefault()) {
			final Data daten = provided.get(info);
			if (daten != null) {
				result = new Parameter(info);
				result.setDataWithTime(new DataWithTime(daten, System.currentTimeMillis()));
				result.setQuelle(info.getObjekt(), QuellTyp.DEFAULT_OBJEKT);
			}
		}

		return result;
	}

	/**
	 * ermittelt, ob die Parameter für die übergebene Spezifikation
	 * bereitgestellt werden.
	 *
	 * @param info
	 *            die Spezifikation
	 * @return <i>true</i>, wenn die Parameter bereitgestellt werden
	 */
	boolean isConnected(final ParameterInfo info) {
		return provided.containsKey(info);
	}

	/**
	 * prüft, ob die durch die übergebene Parameterspezifikation angegebenen
	 * Daten versorgt werden.<br>
	 * Die Parameter für die Parametrierung selbst werden vom Parametermodul
	 * direkt verwaltet und werden deshalb hier ausgeblendet.
	 *
	 * @param info
	 *            die Parameterspezifikation
	 * @return <i>true</i>, wenn die Parameter an- bze. abgemeldet werden können
	 */
	private boolean isProvidedIdentifikation(final ParameterInfo info) {
		boolean result = true;
		if (isParametrierungsParameter(info)) {
			result = false;
		}

		return result;
	}

	/**
	 * prüft, ob die durch die übergebene Parameterspezifikation angegebenen
	 * Daten den Parameterdatensatz der Parametrierung darstellen.
	 *
	 * @param info
	 *            die Parameterspezifikation
	 * @return <i>true</i>, wenn es sich um den Parametrierungsdatensatz handelt
	 */
	boolean isParametrierungsParameter(final ParameterInfo info) {
		boolean result = false;
		if (info.getObjekt().equals(parametrierung.getParameterObjekt())) {
			if ("atg.parametrierung".equals(info.getAtg().getPid())) {
				result = true;
			}
		}

		return result;
	}

	@Override
	public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
		return false;
	}

	/**
	 * nimmt eine neue Parametervorgabe entgegen und trägt sie in die
	 * Warteschlange zur Speicherung und Veröffentlichung ein.
	 *
	 * @param vorgabe
	 *            die neue Vorgabe
	 */
	private void neueVorgabe(final ResultData vorgabe) {
		synchronized (vorgaben) {
			for (final ResultData result : vorgaben) {
				if (result.getObject().equals(vorgabe.getObject())) {
					if (result.getDataDescription().getAttributeGroup()
							.equals(vorgabe.getDataDescription().getAttributeGroup())) {
						vorgaben.remove(result);
						break;
					}
				}
			}
			vorgaben.add(vorgabe);
		}
		synchronized (senderMutex) {
			senderMutex.notify();
		}
	}

	@Override
	public void publiziereHierarchieParameter(final HierarchieObjekt hierarchieObjekt, final Parameter parameter) {

		try {
			if (parameter.getData() == null) {
				parametrierung.getPersistenz().entferne(parameter.getInfo());
			} else {
				parametrierung.getPersistenz().schreibe(parameter);
			}
			final SystemObjectType typ = parameter.getTyp();

			HierarchieObjekt[] sendeListe = new HierarchieObjekt[] { hierarchieObjekt };

			while (sendeListe.length > 0) {
				final Collection<HierarchieObjekt> nachfolger = new ArrayList<>();
				for (final HierarchieObjekt objekt : sendeListe) {
					if (objekt.getObjekt().isOfType(typ)) {
						publiziereParameter(
								new ParameterInfo(objekt.getObjekt(), parameter.getAtg(), parameter.getSim()));
					} else {
						nachfolger.addAll(objekt.getNachFolgerListe());
					}
				}

				sendeListe = nachfolger.toArray(new HierarchieObjekt[nachfolger.size()]);
			}
		} catch (final DatenSchreibenException e) {
			LogTools.log(ParameterProvider.LOGGER, ParamMessage.ErrorSchreibenDatensatz, e.getMessage());
		}
	}

	@Override
	public void publiziereLokalenParameter(final Parameter parameter) {
		try {
			parametrierung.getPersistenz().schreibe(parameter);
			publiziereParameter(parameter.getInfo());
		} catch (final DatenSchreibenException e) {
			LogTools.log(ParameterProvider.LOGGER, ParamMessage.ErrorSchreibenDatensatz, e.getMessage());
		}
	}

	/**
	 * veröffentlicht die Daten für die übergebene Parameterspezifikation.
	 *
	 * @param info
	 *            die Parameterspezifikation
	 */
	private synchronized void publiziereParameter(final ParameterInfo info) {

		if (provided.containsKey(info) || isParametrierungsParameter(info)) {
			final Parameter parameter = parametrierung.ermittleParameter(info, true);
			final Data data = parameter.getData();
			if (data != null && data.isDefined()) {
				final ResultData daten = new ResultData(info.getObjekt(),
						new DataDescription(info.getAtg(), sollAsp, info.getSim()), parameter.getZeitpunkt(),
						parameter.getData());
				ParameterProvider.LOGGER.finer("Publiziere:", daten);
				try {
					parametrierung.getVerbindung().sendData(daten);
				} catch (final SendSubscriptionNotConfirmed e) {
					// sollte nicht vorkommen, da die Parametrierung hier als
					// Quelle fungiert
					throw new RuntimeException("Publizieren der Daten nicht möglich: " + info, e);
				} catch (final DataNotSubscribedException e) {
					LogTools.log(ParameterProvider.LOGGER, ParamMessage.ErrorSendeAnmeldungWithException, info,
							e.getLocalizedMessage());
				}
			} else {
				LogTools.log(ParameterProvider.LOGGER, ParamMessage.ErrorKeineParameterDaten, info.getObjekt(),
						info.getAtg(), info.getSim());
			}
		} else {
			ParameterProvider.LOGGER.info("Parameter werden nicht bereitgestellt: " + info);
		}
	}

	@Override
	public void publiziereTypParameter(final SystemObjectType typ, final Parameter parameter) {

		try {
			parametrierung.getPersistenz().schreibe(parameter);
			for (final ParameterInfo info : provided.keySet()) {
				if (info.getObjekt().isOfType(typ)) {
					if (info.getAtg().equals(parameter.getAtg())) {
						publiziereParameter(new ParameterInfo(info.getObjekt(), info.getAtg(), info.getSim()));
					}
				}
			}
		} catch (final DatenSchreibenException e) {
			LogTools.log(ParameterProvider.LOGGER, ParamMessage.ErrorSchreibenDatensatz, e.getMessage());
		}
	}

	@Override
	public void run() {
		while (true) {
			ResultData vorgabe = null;
			synchronized (vorgaben) {
				vorgabe = vorgaben.poll();
			}
			if (vorgabe == null) {
				try {
					synchronized (senderMutex) {
						senderMutex.wait(30000);
					}
				} catch (final InterruptedException e) {
					ParameterProvider.LOGGER.warning("Parameterprovider wurde unerwartet unterbrochen", e);
				}
			} else {
				final Parameter parameter = new Parameter(
						new ParameterInfo(vorgabe.getObject(), vorgabe.getDataDescription().getAttributeGroup(),
								vorgabe.getDataDescription().getSimulationVariant()));
				parameter.setDataWithTime(new DataWithTime(vorgabe.getData(), vorgabe.getDataTime()));
				try {
					parametrierung.getPersistenz().schreibe(parameter);
					ParameterInfo publikationsInfo = null;
					publikationsInfo = new ParameterInfo(vorgabe.getObject(),
							vorgabe.getDataDescription().getAttributeGroup(),
							vorgabe.getDataDescription().getSimulationVariant());
					publiziereParameter(publikationsInfo);
				} catch (final DatenSchreibenException e) {
					LogTools.log(ParameterProvider.LOGGER, ParamMessage.ErrorSchreibenDatensatz, e.getMessage());
				}
			}
		}
	}

	@Override
	public void update(final ResultData[] results) {
		for (final ResultData result : results) {
			final Data data = result.getData();
			if (data != null) {
				neueVorgabe(result);
			} else {
				ParameterProvider.LOGGER.finer("Leerer Datensatz beim Schreiben eines Parameters ignoriert");
				// leere Vorgabe-Datensätze werden ignoriert. Das Löschen von
				// Parametern kann nur über die Parametrierungsschnittstelle
				// erfolgen.
			}
		}
	}

	@Override
	public boolean isProvided(final ParameterInfo info) {
		boolean result = provided.containsKey(info);
		if (!result) {
			final SystemObject objekt = info.getObjekt();
			if (objekt != null) {
				result = parametrierung.getHierarchieManager().getHierarchieObjekt(objekt.getId()) != null;
			}
		}
		if (!result) {
			result = info.getObjekt() instanceof SystemObjectType;
		}
		if (!result) {
			result = "atg.parametrierung".equals(info.getAtg().getPid());
		}
		return result;
	}

	/** bereinigt die Persistent von nicht mehr verfügbaren Objektparametern. */
	public void cleanup() {
		parametrierung.getPersistenz().cleanup(this);
	}
}
