/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.param;

import java.text.DateFormat;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.puk.param.anfrage.ParameterAnfrageBeantworter;
import de.bsvrz.puk.param.hierarchie.ParametrierungsHierarchieManager;
import de.bsvrz.puk.param.lib.DatenSchreibenException;
import de.bsvrz.puk.param.lib.Parameter;
import de.bsvrz.puk.param.lib.ParameterInfo;
import de.bsvrz.puk.param.lib.daten.DataWithTime;
import de.bsvrz.sys.funclib.bitctrl.daf.DavTools;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.dataIdentificationSettings.DataIdentification;
import de.bsvrz.sys.funclib.dataIdentificationSettings.EndOfSettingsListener;
import de.bsvrz.sys.funclib.dataIdentificationSettings.SettingsManager;
import de.bsvrz.sys.funclib.dataIdentificationSettings.UpdateListener;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Realisierung des eigentlichen Parametrierungsmoduls.<br>
 * Das Modul verarbeitet den Datensatz zur Parametrierung der
 * Parametrierungsapplikation selbst, außerdem wird die Schnittstelle
 * "atg.parameterSchnittstelle" bedient, um Anfragen von Clientapplikationen zu
 * Parametern und der Objekthierarchie zu beantworten.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class Parametrierung implements ClientReceiverInterface, ClientSenderInterface, UpdateListener,
		EndOfSettingsListener, ParametrierungsModul {

	/** der Logger für Logmeldungen. */
	private static final Debug LOGGER = Debug.getLogger();

	/** verwaltet die Standardparameter für alle Parameterspezifikationen. */
	private final DefaultParameterProvider defaultProvider;

	/**
	 * stellt die Parameter für Objekte zur Verfügung und verarbeitet
	 * Parametervorgaben über den Aspekt "asp.parameterVorgabe".
	 */
	private final ParameterProvider provider;

	/** ermittelt die Parameter für eine angefragte Parameterspezifikation. */
	private final ParameterErmittler ermittler;

	/** die verwendete Parametrierungsapplikation. */
	private final ParamApp param;

	/** die Datenbeschreibung für die Vorgabe des Parametrierungsdatensatzes. */
	private final DataDescription vorgabeDesc;

	/**
	 * die Datenbeschreibung für die Veröffentlichung des
	 * Parametrierungsdatensatzes.
	 */
	private final DataDescription sollDesc;

	/**
	 * die Datenbeschreibung für die Antwort auf Anfragen an die Parametrierung.
	 */
	private final DataDescription antwortDesc;

	/** das Verwaltungsmodul für die Hierarchiedefinitionen. */
	private final ParametrierungsHierarchieManager hierarchieManager;

	/** das Persistenzmodul. */
	private final Persistenz persistenz;

	/**
	 * das Datenverteiler-Systemobjekt, das die Parametrierungsapplikation in
	 * der Datenverteiler-Konfiguration repräsentiert.
	 */
	private final SystemObject parameterObjekt;

	/**
	 * die Datenbeschreibung zum Empfang von Anfragen über die Schnittstelle
	 * "atg.parametrierungsAnfrage" oder "atg.parametrierungSchnittstelle".
	 */
	private final DataDescription anfrageDesc;

	/**
	 * Merker, ob die Standardparameter für Objekte über den veralteten Aspekt
	 * "asp.parameterDefault" ausgelesen werden.
	 */
	private final boolean oldDefault;

	/**
	 * die Liste der bei einer Parameterdatensatzänderung empfangenen neuen
	 * Spezifikationen.
	 */
	private final Set<ParameterInfo> addedSettings = new HashSet<>();

	/**
	 * die Liste der bei einer Parameterdatensatzänderung empfangenen entfernten
	 * Spezifikationen.
	 */
	private final Set<ParameterInfo> removedSettings = new HashSet<>();

	/** interner Zähler für den Empfang von Datenidentifikationsmeldungen. */
	private long settingsStart;

	/**
	 * markiert, ob die initiale Bereinigung der Datenbankdaten (nach erstem
	 * Abschluß des SettingsManagers) bereits ausgeführt wurde.
	 */
	private boolean initialCleanupReady;

	/**
	 * Konstruktor, erzeugt eine Instanz des Parametrierungsmoduls.
	 *
	 * @param param
	 *            die verwendete Parametrierungsapplikation
	 * @param parametrierungsObjekt
	 *            das Objekt der Parametrierung
	 * @param persistenz
	 *            die verwendete Persistenz
	 * @param oldDefault
	 *            Zugriff auf Standardparameter über den Aspekt
	 *            "asp.parameterDefault" ermöglichen
	 */
	Parametrierung(final ParamApp param, final SystemObject parametrierungsObjekt, final Persistenz persistenz,
			final boolean oldDefault) {
		this.param = param;
		this.persistenz = persistenz;
		this.oldDefault = oldDefault;

		final ClientDavInterface verbindung = param.getDavVerbindung();

		Parametrierung.LOGGER.info("Initialisiere Standardparameter-Provider");
		defaultProvider = new DefaultParameterProvider(verbindung);

		Parametrierung.LOGGER.info("Initialisiere Hierarchie-Manager");
		hierarchieManager = new ParametrierungsHierarchieManager(this);

		Parametrierung.LOGGER.info("Ermittle Datenverteilerobjekte für die Parameterverwaltung");
		final DataModel model = verbindung.getDataModel();
		parameterObjekt = parametrierungsObjekt;
		AttributeGroup atg = model.getAttributeGroup("atg.parametrierung");

		vorgabeDesc = new DataDescription(atg, model.getAspect("asp.parameterVorgabe"),
				verbindung.getClientDavParameters().getSimulationVariant());
		sollDesc = new DataDescription(atg, model.getAspect("asp.parameterSoll"),
				verbindung.getClientDavParameters().getSimulationVariant());

		atg = model.getAttributeGroup("atg.parametrierungsAnfrage");
		if (atg == null) {
			atg = model.getAttributeGroup("atg.parametrierungSchnittstelle");
		}
		anfrageDesc = new DataDescription(atg, model.getAspect("asp.anfrage"),
				verbindung.getClientDavParameters().getSimulationVariant());

		atg = model.getAttributeGroup("atg.parametrierungsAntwort");
		if (atg == null) {
			atg = model.getAttributeGroup("atg.parametrierungSchnittstelle");
		}
		antwortDesc = new DataDescription(atg, model.getAspect("asp.antwort"),
				verbindung.getClientDavParameters().getSimulationVariant());

		Parametrierung.LOGGER.info("Initialisiere Parameterdaten-Provider");
		provider = new ParameterProvider(this);
		provider.start();

		final DataIdentification identifikation = new DataIdentification(parameterObjekt, sollDesc);

		Parametrierung.LOGGER.info("Initialisiere Settingsmanager für: ", identifikation);
		final SettingsManager settingsManager = new SettingsManager(verbindung, identifikation);
		settingsManager.addUpdateListener(this);
		settingsManager.addEndOfSettingsListener(this);
		settingsManager.start();

		DavTools.validiereSimulationsVariante(vorgabeDesc.getSimulationVariant());
		verbindung.subscribeReceiver(this, parameterObjekt, vorgabeDesc, ReceiveOptions.normal(), ReceiverRole.drain());

		DavTools.validiereSimulationsVariante(anfrageDesc.getSimulationVariant());
		verbindung.subscribeReceiver(this, parameterObjekt, anfrageDesc, ReceiveOptions.normal(), ReceiverRole.drain());

		try {
			DavTools.validiereSimulationsVariante(sollDesc.getSimulationVariant());

			final Parameter parameter = persistenz
					.lies(new ParameterInfo(parameterObjekt, sollDesc.getAttributeGroup(), (short) 0));
			long zeitPunkt = parameter.getZeitpunkt();
			if (zeitPunkt <= 0) {
				zeitPunkt = verbindung.getTime();
			}
			verbindung.subscribeSource(this, new ResultData(parameterObjekt, sollDesc, zeitPunkt, parameter.getData()));

		} catch (final OneSubscriptionPerSendData e) {
			throw new RuntimeException(e);
		}

		ermittler = new ParameterErmittler(this);
	}

	@Override
	public void close() {
		if (persistenz != null) {
			persistenz.close();
		}
	}

	@Override
	public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
		Parametrierung.LOGGER.info("Unerwarteter Aufruf von dataRequest",
				new Object[] { object, dataDescription, state });
		// die Funktion wird nicht verwendet, da keine Sendesteuerung
		// unterstützt wird
	}

	@Override
	public Parameter ermittleParameter(final ParameterInfo info, final boolean datenLiefern) {
		Parameter result = null;

		if (!info.getObjekt().isOfType(info.getTyp()) || provider.isConnected(info)
				|| provider.isParametrierungsParameter(info)) {
			if (ermittler != null) {
				result = ermittler.ermittleParameter(info);
			}
		}

		if (result != null) {
			if (result.getData() != null) {
				result.getQuelle().setDataAvailable(true);
			}

			if (!datenLiefern) {
				result.setDataWithTime((DataWithTime) null);
			}
		}

		return result;
	}

	/**
	 * liefert den Standardwert für die übergeben Parameterspezifikation.
	 *
	 * @param info
	 *            die Spezifikation
	 * @return die Daten
	 */
	DataWithTime getDefaultParameterDaten(final ParameterInfo info) {

		DataWithTime result = null;
		Parameter parameter = null;

		if (isOldDefault()) {
			parameter = provider.getOldStyleDefaultDaten(info);
		}

		if (parameter == null) {
			parameter = defaultProvider.getDefaultParameter(info);
		}

		if (parameter != null) {
			result = parameter.getDataWithTime();
		}

		return result;
	}

	@Override
	public ParametrierungsHierarchieManager getHierarchieManager() {
		return hierarchieManager;
	}

	@Override
	public SystemObject getParameterObjekt() {
		return parameterObjekt;
	}

	@Override
	public Persistenz getPersistenz() {
		return persistenz;
	}

	@Override
	public Provider getProvider() {
		return provider;
	}

	@Override
	public ClientDavInterface getVerbindung() {
		return param.getDavVerbindung();
	}

	/**
	 * Alle nicht mehr benötigten Anmeldungen werden entfernt, neue Anmeldungen
	 * werden umgesetzt.
	 */
	@Override
	public void inform() {
		synchronized (addedSettings) {
			synchronized (removedSettings) {
				Parametrierung.LOGGER.finer("SettingsManager fertig, aktualisiere Hierarchie");
				hierarchieManager.aktualisiereObjekte();
				Parametrierung.LOGGER.finer("Hierarchie aktualisiert");
				Parametrierung.LOGGER.info("SettingsStart: "
						+ DateFormat.getDateTimeInstance().format(new Date(settingsStart)) + " Parametrieren: "
						+ addedSettings.size() + " Parametrieren beenden: " + removedSettings.size());

				for (final ParameterInfo info : removedSettings) {
					provider.disconnect(info, true);
				}
				removedSettings.clear();

				for (final ParameterInfo info : addedSettings) {
					provider.connect(info);
				}
				addedSettings.clear();

				if (settingsStart == 0) {
					Parametrierung.LOGGER.info("SettingsManager meldet keine Änderungen");
				} else {
					Parametrierung.LOGGER.info("Parameterspezifikationen angemeldet in "
							+ (System.currentTimeMillis() - settingsStart) + " Millisekunden");
					settingsStart = 0;
				}

				if (!initialCleanupReady) {
					provider.cleanup();
					initialCleanupReady = true;
				}
			}
		}

		if (param != null) {
			param.sendReadyMessage();
		}
	}

	@Override
	public void initialisiereSimulationsParameter(final short simulationsVariante,
			final Set<ParameterInfo> anmeldungen) {
		persistenz.entferneSimulation(simulationsVariante);
		persistenz.erzeugeSimulation(simulationsVariante);
	}

	/**
	 * ermittelt, ob Standardwerte über den veralteten Aspekt
	 * "asp.parameterDefault" ausgelesen werden sollen.
	 *
	 * @return <i>true</i>, wenn sie ausgelesen werden sollen
	 */
	boolean isOldDefault() {
		return oldDefault;
	}

	@Override
	public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
		return false;
	}

	/**
	 * die Funktion ermittelt, ob die übergebenen Daten eine Parametrierung
	 * fordern oder nicht.
	 *
	 * @param settings
	 *            die Daten
	 * @return true, wenn die Parametrierung gefordert ist.
	 */
	private boolean parametrierenGefordert(final Data settings) {
		boolean result = false;
		if (settings != null) {
			if (settings.getUnscaledValue("Parametrieren").intValue() != 0) {
				result = true;
			}
		}
		return result;
	}

	/**
	 * Die Funktion sammelt die neuen Settingseinstellungen, die
	 * Veröffentlichung erfolgt mit dem Aufruf des {@link EndOfSettingsListener}
	 * .
	 */
	@Override
	public void update(final DataIdentification dataIdentification, final Data oldSettings, final Data newSettings) {

		if (settingsStart == 0) {
			settingsStart = System.currentTimeMillis();
		}

		final boolean parametrieren = parametrierenGefordert(newSettings);

		synchronized (addedSettings) {
			synchronized (removedSettings) {
				Parametrierung.LOGGER.fine("Update für", dataIdentification);
				if (parametrieren) {
					Parametrierung.LOGGER.fine("Anmeldung", dataIdentification);
					addedSettings.add(new ParameterInfo(dataIdentification.getObject(),
							dataIdentification.getDataDescription().getAttributeGroup(),
							dataIdentification.getDataDescription().getSimulationVariant()));
				} else {
					Parametrierung.LOGGER.fine("Abmeldung", dataIdentification);
					removedSettings.add(new ParameterInfo(dataIdentification.getObject(),
							dataIdentification.getDataDescription().getAttributeGroup(),
							dataIdentification.getDataDescription().getSimulationVariant()));
				}
			}
		}
	}

	@Override
	public void update(final ResultData[] results) {
		for (final ResultData result : results) {
			try {
				if (result.getData() != null) {
					if (sollDesc.getAttributeGroup().equals(result.getDataDescription().getAttributeGroup())) {
						final Parameter parameter = new Parameter(new ParameterInfo(parameterObjekt,
								vorgabeDesc.getAttributeGroup(), vorgabeDesc.getSimulationVariant()));
						parameter.setDataWithTime(new DataWithTime(result.getData(), result.getDataTime()));
						try {
							persistenz.schreibe(parameter);
							getVerbindung().sendData(new ResultData(result.getObject(), sollDesc, result.getDataTime(),
									result.getData()));
						} catch (final DataNotSubscribedException e) {
							// darf nicht passieren
							throw new RuntimeException(e);
						} catch (final SendSubscriptionNotConfirmed e) {
							// darf nicht passieren
							throw new RuntimeException(e);
						} catch (final DatenSchreibenException e) {
							LogTools.log(Parametrierung.LOGGER, ParamMessage.ErrorSchreibenDatensatz, e.getMessage());
						}
					} else if (anfrageDesc.getAttributeGroup()
							.equals(result.getDataDescription().getAttributeGroup())) {
						ParameterAnfrageBeantworter.erzeugeBeantworter(this, result.getData()).start();
					}
				}
			} catch (final IllegalStateException ex) {
				Parametrierung.LOGGER.warning("Fehler beim Datenempfang", ex);
			}
		}
	}

	/**
	 * liefert die Databeschreibung für die Antworten der Parametrierung über
	 * die Anfrageschnittstelle.
	 *
	 * @return die Datenbeschreibung
	 */
	public DataDescription getAntwortDesc() {
		return antwortDesc;
	}
}
