/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.param;

import com.bitctrl.Constants;

import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList.Argument;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Parameter für das Persitenzmodul der Parametrierung.
 *
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
final class PersistenzParameter {

	/** globale Instanz der Persistenzparameter. */
	private static final PersistenzParameter INSTANZ = new PersistenzParameter();

	/**
	 * der Argumentname zur Angabe des Startparameters für das Persistenzmodul.
	 */
	static final String PARAMETERNAME_PERSISTENZ = "-persistenz";

	/** der Argumentname zur Angabe der Cachegröße der Persistenz. */
	protected static final String PARAMETERNAME_CACHE_ANZAHL = "-cacheGroesse";

	/** der Argumentname für den "Aufräumzyklus" der Persistenz in Minuten. */
	protected static final String PARAMETERNAME_BEREINIGUNG_ZYKLUS = "-bereinigungsZyklus";

	/** Standardwert für die Anzahl der Einträge im Persistenzcache. */
	private static final int MAXIMALE_ANZAHL_CACHE_DEFAULT = 200000;

	/** Standardwert für die Anzahl der Einträge im Persistenzcache. */
	private static final long STANDARD_BEREINIGUNGS_ZYKLUS = Constants.MILLIS_PER_HOUR;

	/**
	 * liefert die globale Instanz des Objekts zum Speichern der Parameter für
	 * das Persistenzmodul.
	 *
	 * @return die Instanz
	 */
	static PersistenzParameter getInstanz() {
		return PersistenzParameter.INSTANZ;
	}

	/** Name des Wurzelverzeichnisses für die Persistenzdaten. */
	private String wurzel = "../../parameter";

	/** Anzahl der Einträge im Persistenzcache. */
	private int cacheAnzahl = PersistenzParameter.MAXIMALE_ANZAHL_CACHE_DEFAULT;

	/** Zyklus für die Bereinigung der Persitenzdaten in Millisekunden. */
	private long bereinigungsZyklus = PersistenzParameter.STANDARD_BEREINIGUNGS_ZYKLUS;

	/**
	 * Privater Konstruktor, der das Anlegen zusätzlcher Objekte dieser Klasse
	 * verhindern soll.
	 */
	private PersistenzParameter() {
		super();
	}

	/**
	 * liefert den Bereinigungszyklus in Millisekunden.
	 *
	 * @return den Wert ino Millisekunden
	 */
	long getBereinigungsZyklus() {
		if (bereinigungsZyklus != 0 && bereinigungsZyklus <= Constants.MILLIS_PER_MINUTE) {
			bereinigungsZyklus = PersistenzParameter.STANDARD_BEREINIGUNGS_ZYKLUS;
		}
		return bereinigungsZyklus;
	}

	/**
	 * liefert die Anzahl der Einträge, die im Cache der Persistemz abgelegt
	 * werden sollen.
	 *
	 * @return die Anzahl
	 */
	protected int getCacheAnzahl() {
		return cacheAnzahl;
	}

	/**
	 * liefert den Namen des Wurzelverzeichnisses für die Daten des
	 * Persistemzmoduls.
	 *
	 * @return der Name des Verzeichnisses
	 */
	protected String getPersistenzWurzel() {
		return wurzel;
	}

	/**
	 * setzt einen Parameter.
	 *
	 * @param argument
	 *            das Argument, das den Parameter beschreibt
	 */
	void setzeParameter(final Argument argument) {
		if (argument.getName().equals(PersistenzParameter.PARAMETERNAME_PERSISTENZ)) {
			wurzel = argument.getValue();
		} else if (argument.getName().equals(PersistenzParameter.PARAMETERNAME_CACHE_ANZAHL)) {
			cacheAnzahl = Integer.parseInt(argument.getValue());
			if (cacheAnzahl <= 0) {
				cacheAnzahl = PersistenzParameter.MAXIMALE_ANZAHL_CACHE_DEFAULT;
			}
		} else if (argument.getName().equals(PersistenzParameter.PARAMETERNAME_BEREINIGUNG_ZYKLUS)) {
			bereinigungsZyklus = Integer.parseInt(argument.getValue()) * Constants.MILLIS_PER_MINUTE;
			Debug.getLogger().config("Bereinigungszyklus: " + bereinigungsZyklus + " Millisekunden");
		}
	}
}
