/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.simulation;

import java.util.HashMap;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.puk.param.param.ParamMessage;
import de.bsvrz.puk.param.param.ParametrierungsModul;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.DavTools;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * die Klasse, die die überwachung einer Simulation implementiert.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
class SimulationsHandler implements SimulationsStatusHandler, ClientSenderInterface {

	/**
	 * Name des Attributs zur übertragung des Bereitschaftsstatus der
	 * Parametrierung.
	 */
	private static final String ATT_NAME_PARAMETRIERUNG_BEREIT = "ParametrierungBereit";

	/**
	 * Logger für Log-Meldungen.
	 */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * die Liste aller verwalteten Simulationen.
	 */
	private static final Map<Short, SimulationsHandler> HANDLER = new HashMap<>();

	/**
	 * die Datenbeschreibung zur Veröffentlichung des Status der Parametrierung
	 * in Bezug zu einer Simulation.
	 */
	private static DataDescription parameterStatusDesc;

	/**
	 * liefert den mit dem übergebenen Systemobjekt beschriebenen und
	 * registrierten SimulationsHandler.
	 *
	 * @param simulationsObjekt
	 *            das Objekt, das die Simulations definiert
	 * @return den Handler oder <i>null</i>, wenn kein solcher registriert ist
	 */
	static SimulationsHandler getHandler(final SystemObject simulationsObjekt) {
		SimulationsHandler result = null;
		for (final SimulationsHandler handler : SimulationsHandler.HANDLER.values()) {
			if (handler.simulationsObjekt.equals(simulationsObjekt)) {
				result = handler;
				break;
			}
		}
		return result;
	}

	/**
	 * der Beobachter zur überwachung des Status der Simulation.
	 */
	private final SimulationsStatusObserver observer;

	/**
	 * das Systemobjekt, das die Simulation innerhlab der
	 * Datenverteilerkonfiguration repräsentiert.
	 */
	private final SystemObject simulationsObjekt;

	/**
	 * die Simulationsvariante.
	 */
	private final short simulationsVariante;

	/**
	 * die Simulationsverwaltung, die die Simulationen verwaltet.
	 */
	private final SimulationsVerwaltung simulationsVerwaltung;

	/**
	 * der Manager für die Interpretation der bereitzustellenden Parameter.
	 */
	private final SimulationsSettingsManager settingsManager;

	/**
	 * Marker für den Aktivitätszustand der Simulation innerhalb der
	 * Parametrierung.
	 */
	private boolean aktiv;

	/**
	 * die Simulationsstrecke, die der Simulation zugeordnet its.
	 */
	private final SystemObject simulationsStrecke;

	/**
	 * konstruiert einen SimulationsHandler für das übergebene Simlationsobjekt
	 * und die Parametrierung.
	 *
	 * @param simulation
	 *            die Simulationsverwaltung, die alle Simulationen verwaltet
	 * @param simulationsObjekt
	 *            das Objekt, das die Simulation repräsentiert
	 * @throws InvalidSimulationException
	 *             die Simulations konnte auf Grund ungültiger
	 *             Konfigurationsdaten nicht angelegt werden
	 */
	SimulationsHandler(final SimulationsVerwaltung simulation, final SystemObject simulationsObjekt)
			throws InvalidSimulationException {

		this.simulationsVerwaltung = simulation;
		this.simulationsObjekt = simulationsObjekt;

		final DataModel model = simulation.getVerbindung().getDataModel();

		if (SimulationsHandler.parameterStatusDesc == null) {
			SimulationsHandler.parameterStatusDesc = new DataDescription(
					model.getAttributeGroup("atg.simulationsStatusParametrierung"), model.getAspect("asp.zustand"),
					(short) 0);
		}

		final Data daten = simulationsObjekt
				.getConfigurationData(model.getAttributeGroup("atg.simulationsEigenschaften"));

		simulationsVariante = daten.getUnscaledValue("SimulationsVariante").shortValue();

		if (SimulationsHandler.HANDLER.containsKey(simulationsVariante)) {
			throw new InvalidSimulationException(
					"Es existiert bereits eine Simulation mit der Variante: " + simulationsVariante);
		}

		simulationsStrecke = daten.getReferenceValue("SimulationsStreckenReferenz").getSystemObject();
		if (simulationsStrecke == null) {
			throw new InvalidSimulationException(
					"Die Simulation hat keine zugeordnete Simulationsstrecke: " + simulationsObjekt);
		}

		try {
			simulation.getVerbindung().subscribeSender(this, simulationsObjekt, SimulationsHandler.parameterStatusDesc,
					SenderRole.source());
		} catch (final OneSubscriptionPerSendData e) {
			// darf nicht auftreten, es sei denn eine andere Parametrierung
			// läuft bereits
			throw new RuntimeException(
					"Die Anmeldung als Quelle für die Bereitmeldung der Parametrierung ist nicht möglich: "
							+ simulationsObjekt,
					e);
		}

		settingsManager = new SimulationsSettingsManager(this);

		observer = new SimulationsStatusObserver(simulation.getVerbindung(), simulationsObjekt);
		observer.addSimulationsStatusHandler(this);
		observer.start();
	}

	/**
	 * setzt den Aktivierungsstatus einer Simulation und veröffentlicht diesen.
	 *
	 * @param status
	 *            der Status
	 */
	void aktiviereSimulation(final boolean status) {
		aktiv = status;
		bereitschaftMelden();
	}

	/**
	 * die Funktion veröffentlicht den Bereitschaftsstatus der Parametrierung
	 * bezüglich der Simulation.
	 */
	private void bereitschaftMelden() {

		final Data daten = simulationsVerwaltung.getVerbindung()
				.createData(SimulationsHandler.parameterStatusDesc.getAttributeGroup());
		daten.getUnscaledValue(SimulationsHandler.ATT_NAME_PARAMETRIERUNG_BEREIT)
				.set(DavTools.intWertVonBoolean(aktiv));
		try {
			simulationsVerwaltung.getVerbindung().sendData(new ResultData(simulationsObjekt,
					SimulationsHandler.parameterStatusDesc, System.currentTimeMillis(), daten));
		} catch (final DataNotSubscribedException e) {
			// sollte nocht auftreten
			LogTools.log(SimulationsHandler.LOGGER, new BetriebsmeldungDaten(simulationsObjekt),
					ParamMessage.ErrorBereitschaftSimulation, simulationsObjekt, e.getMessage());
		} catch (final SendSubscriptionNotConfirmed e) {
			// sollte nocht auftreten
			LogTools.log(SimulationsHandler.LOGGER, new BetriebsmeldungDaten(simulationsObjekt),
					ParamMessage.ErrorBereitschaftSimulation, simulationsObjekt, e.getMessage());
		}
	}

	@Override
	public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
		SimulationsHandler.LOGGER.info("Unerwarteter Sendesteuerungsmeldung: ",
				new Object[] { object, dataDescription, state });
	}

	/**
	 * entfernt die gepeicherten Parameterdatensätze.
	 */
	private void entfernePersistenteSimulationsParameter() {
		simulationsVerwaltung.getParametrierung().getPersistenz().entferneSimulation(simulationsVariante);
	}

	/**
	 * liefert die Parametrierung.
	 *
	 * @return die Parametrierung
	 */
	ParametrierungsModul getParametrierung() {
		return simulationsVerwaltung.getParametrierung();
	}

	/**
	 * liefert das Objekt, das die Simulation innerhalb des Datenverteilers
	 * repräsentiert.
	 *
	 * @return simulationsObjekt
	 */
	SystemObject getSimulationsObjekt() {
		return simulationsObjekt;
	}

	/**
	 * liefert die verwendete Simulationsstrecke.
	 *
	 * @return simulationsStrecke
	 */
	SystemObject getSimulationsStrecke() {
		return simulationsStrecke;
	}

	/**
	 * liefert die Simulationsvariante der Simulation.
	 *
	 * @return simulationsVariante
	 */
	short getSimulationsVariante() {
		return simulationsVariante;
	}

	/**
	 * liefert die verwendete Datenverteilerverbindung.
	 *
	 * @return die Verbindung
	 */
	ClientDavInterface getVerbindung() {
		return simulationsVerwaltung.getVerbindung();
	}

	/**
	 * ermittelt, ob die Simulation innerhalb der Parametrierung aktiv ist.<br>
	 * Das heißt, es werden können Parameter publiziert werden oder nicht.
	 *
	 * @return den Aktivitätszustand
	 */
	boolean isAktiv() {
		return aktiv;
	}

	@Override
	public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
		return false;
	}

	/**
	 * Die Funktion implementiert die entsprechende Funktion der Schnittstelle
	 * des {@link SimulationsStatusHandler}. Je nach Status der Simulation
	 * werden die entsprechenden Operationen ausgeführt.
	 * <table summary="Statuswerte einer Simulation">
	 * <tr>
	 * <td>Status</td>
	 * <td>Operation</td>
	 * </tr>
	 * <tr>
	 * <td>NEU</td>
	 * <td>die in der Parametrierung vorhandenen Parameterdatensätze für den
	 * Standardbetrieb werden für alle für die Simulation spezifizierten Objekte
	 * übernommen.</td>
	 * </tr>
	 * <tr>
	 * <td>VORSTART</td>
	 * <td>die Parametrierung meldet sich für alle spezifizierten Objekte als
	 * Quelle für den Parameter-Sollzustand und als Senke für Parameter-Vorgaben
	 * an. Nach der Anmeldung meldet die Parametrierung ihre Bereitschaft für
	 * die Simulation.</td>
	 * </tr>
	 * <tr>
	 * <td>START</td>
	 * <td>es erfolgt keine Operation</td>
	 * </tr>
	 * <tr>
	 * <td>PAUSE</td>
	 * <td>es erfolgt keine Operation</td>
	 * </tr>
	 * <tr>
	 * <td>STOP</td>
	 * <td>die Parametrierung meldet ihre Bereitschaft für die Simulation und
	 * alle angemeldeten Parameter ab</td>
	 * </tr>
	 * <tr>
	 * <td>GELOESCHT</td>
	 * <td>die persitent gespeicherten Parameter für die Simulationsvariante
	 * werden entfernt</td>
	 * </tr>
	 * </table>
	 */
	@Override
	public void neuerStatus(final SystemObject simulation, final SimulationsStatus status) {

		// TODO Behandlung des Simulationsstatus überprüfen

		SimulationsHandler.LOGGER.info("Verarbeite neuen Simulationsstatus für " + simulation + " --> " + status);

		if (SimulationsStatus.NEU.equals(status)) {
			aktiviereSimulation(false);
			simulationStatusNeuBehandeln();
		} else if (SimulationsStatus.VORSTART.equals(status)) {
			aktiviereSimulation(true);
		} else if (SimulationsStatus.START.equals(status)) {
			SimulationsHandler.LOGGER.finer("Simulationsstatus: " + simulation + ": " + status);
			// nichts zu tun
		} else if (SimulationsStatus.PAUSE.equals(status)) {
			SimulationsHandler.LOGGER.finer("Simulationsstatus: " + simulation + ": " + status);
			// nichts zu tun
		} else if (SimulationsStatus.STOP.equals(status)) {
			aktiviereSimulation(false);
		} else if (SimulationsStatus.GELOESCHT.equals(status)) {
			simulationEntfernen();
		} else {
			LogTools.log(SimulationsHandler.LOGGER, new BetriebsmeldungDaten(simulationsObjekt),
					ParamMessage.WarningUngueltigerSimulationsZustand, simulationsObjekt, status);
		}
	}

	/**
	 * die Funktion entfernt alle gespeicherten Parameter für die
	 * Simulationsvariante.
	 *
	 */
	void simulationEntfernen() {
		aktiviereSimulation(false);
		observer.removeSimulationsStatusHandler(this);
		observer.anhalten();
		entfernePersistenteSimulationsParameter();
		SimulationsHandler.HANDLER.remove(simulationsVariante);
		settingsManager.anhalten();
	}

	/**
	 * die Funktion kopiert initial die Parameterdatensätze für die
	 * angeforderten Objekte.
	 */
	private void simulationStatusNeuBehandeln() {
		aktiv = false;
		settingsManager.abmelden();
		entfernePersistenteSimulationsParameter();
		settingsManager.anmelden();
	}
}
